import sys
from os import path

import betty
from betty import about, fs
from betty.asyncio import wait_to_thread
from betty.fs import ROOT_DIRECTORY_PATH, FileSystem
from betty.locale import LocalizerRepository

betty_replacements: dict[str, str] = {}

assets = FileSystem()
assets.prepend(fs.ASSETS_DIRECTORY_PATH, "utf-8")
localizers = LocalizerRepository(assets)
for locale in localizers.locales:
    coverage = wait_to_thread(localizers.coverage(locale))
    betty_replacements[f"translation-coverage-{locale}"] = str(
        int(round(100 / (coverage[1] / coverage[0])))
    )

sys.path.insert(0, path.dirname(path.dirname(betty.__file__)))
project = "Betty"
version = wait_to_thread(about.version()) or ""
release = wait_to_thread(about.version_label())
copyright = "Bart Feenstra and contributors"
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "jinja2": ("https://jinja.palletsprojects.com/en/latest/", None),
}
html_favicon = str(
    ROOT_DIRECTORY_PATH / "betty" / "assets" / "public" / "static" / "betty.ico"
)
html_logo = str(
    ROOT_DIRECTORY_PATH / "betty" / "assets" / "public" / "static" / "betty-32x32.png"
)
html_context = {
    "display_github": True,
    "github_user": "bartfeenstra",
    "github_repo": "betty",
    "github_version": "0.3.x",
    "conf_py_path": "/documentation/",
    "betty_replacements": betty_replacements,
}
html_theme = "sphinx_immaterial"
highlight_language = "none"
templates_path = ["_templates"]
extensions = [
    "betty.sphinx.extension.replacements",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.viewcode",
    "sphinx_autodoc_typehints",
    "sphinx_design",
    "sphinx_immaterial",
]
nitpicky = True
modindex_common_prefix = ["betty."]

# sphinx.ext.autodoc configuration.
autodoc_member_order = "alphabetical"

# sphinx_autodoc_typehints configuration.
set_type_checking_flag = True
typehints_fully_qualified = True
always_document_param_types = True



"""
Integrate Webpack with Jinja2.
"""

from jinja2.runtime import Context


def _context_js_entrypoints(context: Context) -> set[str]:
    entrypoints = context.resolve_or_missing("webpack_js_entrypoints")
    if isinstance(entrypoints, set):
        return entrypoints
    raise RuntimeError(
        "No `webpack_js_entrypoints` context variable exists in this Jinja2 template."
    )



import logging
from collections import defaultdict
from collections.abc import Sequence, MutableMapping, AsyncIterator
from contextlib import asynccontextmanager
from datetime import datetime
from typing import Generic, Self, overload, AsyncContextManager, Literal

from betty.cache import (
    CacheItemValueCoT,
    Cache,
    CacheItem,
    CacheItemValueContraT,
    CacheItemValueSetter,
)
from betty.concurrent import _Lock, AsynchronizedLock, MultiLock
from betty.locale import Localizer


class _StaticCacheItem(CacheItem[CacheItemValueCoT], Generic[CacheItemValueCoT]):
    __slots__ = "_value", "_modified"

    def __init__(
        self,
        value: CacheItemValueCoT,
        modified: int | float | None = None,
    ):
        self._value = value
        self._modified = datetime.now().timestamp() if modified is None else modified

    async def value(self) -> CacheItemValueCoT:
        return self._value

    @property
    def modified(self) -> int | float:
        return self._modified


class _CommonCacheBase(Cache[CacheItemValueContraT], Generic[CacheItemValueContraT]):
    def __init__(
        self,
        localizer: Localizer,
        *,
        scopes: Sequence[str] | None = None,
    ):
        self._localizer = localizer
        self._scopes = scopes or ()
        self._scoped_caches: dict[str, Self] = {}
        self._locks: MutableMapping[str, _Lock] = defaultdict(
            AsynchronizedLock.threading
        )
        self._locks_lock = AsynchronizedLock.threading()

    async def _lock(self, cache_item_id: str) -> _Lock:
        async with self._locks_lock:
            return self._locks[cache_item_id]

    def with_scope(self, scope: str) -> Self:
        try:
            return self._scoped_caches[scope]
        except KeyError:
            self._scoped_caches[scope] = self._with_scope(scope)
            return self._scoped_caches[scope]

    def _with_scope(self, scope: str) -> Self:
        raise NotImplementedError

    @asynccontextmanager
    async def get(
        self, cache_item_id: str
    ) -> AsyncIterator[CacheItem[CacheItemValueContraT] | None]:
        async with await self._lock(cache_item_id):
            yield await self._get(cache_item_id)

    async def _get(self, cache_item_id: str) -> CacheItem[CacheItemValueContraT] | None:
        raise NotImplementedError

    async def set(
        self,
        cache_item_id: str,
        value: CacheItemValueContraT,
        *,
        modified: int | float | None = None,
    ) -> None:
        async with await self._lock(cache_item_id):
            await self._set(cache_item_id, value, modified=modified)

    async def _set(
        self,
        cache_item_id: str,
        value: CacheItemValueContraT,
        *,
        modified: int | float | None = None,
    ) -> None:
        raise NotImplementedError

    @overload
    def getset(self, cache_item_id: str) -> AsyncContextManager[
        tuple[
            CacheItem[CacheItemValueContraT] | None,
            CacheItemValueSetter[CacheItemValueContraT],
        ]
    ]:
        pass

    @overload
    def getset(
        self, cache_item_id: str, *, wait: Literal[False] = False
    ) -> AsyncContextManager[
        tuple[
            CacheItem[CacheItemValueContraT] | None,
            CacheItemValueSetter[CacheItemValueContraT] | None,
        ]
    ]:
        pass

    @asynccontextmanager  # type: ignore[misc]
    async def getset(self, cache_item_id: str, *, wait: bool = True) -> AsyncIterator[
        tuple[
            CacheItem[CacheItemValueContraT] | None,
            CacheItemValueSetter[CacheItemValueContraT] | None,
        ]
    ]:
        lock = await self._lock(cache_item_id)
        if await lock.acquire(wait=wait):
            try:

                async def _setter(value: CacheItemValueContraT) -> None:
                    await self._set(cache_item_id, value)

                yield await self._get(cache_item_id), _setter
                return
            finally:
                lock.release()
        yield None, None

    async def delete(self, cache_item_id: str) -> None:
        async with await self._lock(cache_item_id):
            await self._delete(cache_item_id)

    async def _delete(self, cache_item_id: str) -> None:
        raise NotImplementedError

    async def clear(self) -> None:
        async with MultiLock(self._locks_lock, *self._locks.values()):
            await self._clear()
        logger = logging.getLogger(__name__)
        if self._scopes:
            logger.info(
                self._localizer._('"{scope}" cache cleared.').format(
                    scope=".".join(self._scopes)
                )
            )
        else:
            logger.info(self._localizer._("All caches cleared."))

    async def _clear(self) -> None:
        raise NotImplementedError



"""
Provide caching that persists cache items to files.
"""

from __future__ import annotations

import asyncio
import shutil
from collections.abc import Sequence
from contextlib import suppress
from os import utime
from pathlib import Path
from pickle import dumps, loads
from typing import Generic, Self

import aiofiles
from aiofiles.ospath import getmtime

from betty.cache import CacheItem, CacheItemValueContraT, CacheItemValueCoT
from betty.cache._base import _CommonCacheBase
from betty.hashid import hashid
from betty.locale import Localizer


class _FileCacheItem(CacheItem[CacheItemValueCoT], Generic[CacheItemValueCoT]):
    __slots__ = "_modified", "_path"

    def __init__(
        self,
        modified: int | float,
        path: Path,
    ):
        self._modified = modified
        self._path = path

    @property
    def modified(self) -> int | float:
        return self._modified

    async def value(self) -> CacheItemValueCoT:
        async with aiofiles.open(self._path, "rb") as f:
            value_bytes = await f.read()
        return await self._load_value(value_bytes)

    async def _load_value(self, value_bytes: bytes) -> CacheItemValueCoT:
        raise NotImplementedError


class _PickledFileCacheItem(
    _FileCacheItem[CacheItemValueCoT], Generic[CacheItemValueCoT]
):
    async def _load_value(self, value_bytes: bytes) -> CacheItemValueCoT:
        return loads(value_bytes)  # type: ignore[no-any-return]


class _BinaryFileCacheItem(_FileCacheItem[bytes]):
    async def _load_value(self, value_bytes: bytes) -> bytes:
        return value_bytes


class _FileCache(
    _CommonCacheBase[CacheItemValueContraT], Generic[CacheItemValueContraT]
):
    """
    Provide a cache that persists cache items on a file system.
    """

    _cache_item_cls: type[_FileCacheItem[CacheItemValueContraT]]

    def __init__(
        self,
        localizer: Localizer,
        cache_directory_path: Path,
        *,
        scopes: Sequence[str] | None = None,
    ):
        super().__init__(localizer, scopes=scopes)
        self._root_path = cache_directory_path

    def _with_scope(self, scope: str) -> Self:
        return type(self)(
            self._localizer, self._root_path, scopes=(*self._scopes, scope)
        )

    def _cache_item_file_path(self, cache_item_id: str) -> Path:
        return self._path / hashid(cache_item_id)

    def _dump_value(self, value: CacheItemValueContraT) -> bytes:
        raise NotImplementedError

    async def _get(self, cache_item_id: str) -> CacheItem[CacheItemValueContraT] | None:
        try:
            cache_item_file_path = self._cache_item_file_path(cache_item_id)
            return self._cache_item_cls(
                await getmtime(cache_item_file_path),
                cache_item_file_path,
            )
        except OSError:
            return None

    async def _set(
        self,
        cache_item_id: str,
        value: CacheItemValueContraT,
        *,
        modified: int | float | None = None,
    ) -> None:
        value = self._dump_value(value)
        cache_item_file_path = self._cache_item_file_path(cache_item_id)
        try:
            await self._write(cache_item_file_path, value, modified)
        except FileNotFoundError:
            await aiofiles.os.makedirs(cache_item_file_path.parent, exist_ok=True)
            await self._write(cache_item_file_path, value, modified)

    async def _write(
        self,
        cache_item_file_path: Path,
        value: bytes,
        modified: int | float | None = None,
    ) -> None:
        async with aiofiles.open(cache_item_file_path, "wb") as f:
            await f.write(value)
        if modified is not None:
            await asyncio.to_thread(utime, cache_item_file_path, (modified, modified))

    async def _delete(self, cache_item_id: str) -> None:
        with suppress(FileNotFoundError):
            await aiofiles.os.remove(self._cache_item_file_path(cache_item_id))

    async def _clear(self) -> None:
        with suppress(FileNotFoundError):
            await asyncio.to_thread(shutil.rmtree, self._path)

    @property
    def _path(self) -> Path:
        return self._root_path.joinpath(*self._scopes)


class PickledFileCache(
    _FileCache[CacheItemValueContraT], Generic[CacheItemValueContraT]
):
    """
    Provide a cache that pickles values and persists them to files.
    """

    _cache_item_cls = _PickledFileCacheItem

    def _dump_value(self, value: CacheItemValueContraT) -> bytes:
        return dumps(value)


class BinaryFileCache(_FileCache[bytes]):
    """
    Provide a cache that persists bytes values to binary files.
    """

    _cache_item_cls = _BinaryFileCacheItem

    def _dump_value(self, value: bytes) -> bytes:
        return value

    @property
    def path(self) -> Path:
        return self._path

    def cache_item_file_path(self, cache_item_id: str) -> Path:
        return self._cache_item_file_path(cache_item_id)



"""Define Betty's core application functionality."""

from __future__ import annotations

import operator
import weakref
from collections.abc import AsyncIterator
from concurrent.futures import Executor, ProcessPoolExecutor
from contextlib import suppress, asynccontextmanager
from functools import reduce
from graphlib import CycleError, TopologicalSorter
from multiprocessing import get_context
from os import environ
from pathlib import Path
from types import TracebackType
from typing import TYPE_CHECKING, Mapping, Self, Any, final

import aiohttp
from aiofiles.tempfile import TemporaryDirectory

from betty import fs
from betty.app.extension import (
    ListExtensions,
    Extension,
    Extensions,
    build_extension_type_graph,
    CyclicDependencyError,
    ExtensionDispatcher,
    ConfigurableExtension,
)
from betty.asyncio import wait_to_thread
from betty.cache import Cache, FileCache
from betty.cache.file import BinaryFileCache, PickledFileCache
from betty.config import Configurable, FileBasedConfiguration
from betty.dispatch import Dispatcher
from betty.fs import FileSystem, CACHE_DIRECTORY_PATH
from betty.locale import LocalizerRepository, get_data, DEFAULT_LOCALE, Localizer, Str
from betty.model import Entity, EntityTypeProvider
from betty.model.event_type import (
    EventType,
    EventTypeProvider,
    Birth,
    Baptism,
    Adoption,
    Death,
    Funeral,
    Cremation,
    Burial,
    Will,
    Engagement,
    Marriage,
    MarriageAnnouncement,
    Divorce,
    DivorceAnnouncement,
    Residence,
    Immigration,
    Emigration,
    Occupation,
    Retirement,
    Correspondence,
    Confirmation,
)
from betty.project import Project
from betty.render import Renderer, SequentialRenderer
from betty.serde.dump import minimize, void_none, Dump, VoidableDump
from betty.serde.load import (
    AssertionFailed,
    Fields,
    Assertions,
    OptionalField,
    Asserter,
)
from betty.warnings import deprecate

if TYPE_CHECKING:
    from betty.jinja2 import Environment
    from betty.serve import Server
    from betty.url import StaticUrlGenerator, LocalizedUrlGenerator

CONFIGURATION_DIRECTORY_PATH = fs.HOME_DIRECTORY_PATH / "configuration"


class _AppExtensions(ListExtensions):
    def __init__(self):
        super().__init__([])

    def _update(self, extensions: list[list[Extension]]) -> None:
        self._extensions = extensions


class AppConfiguration(FileBasedConfiguration):
    def __init__(
        self,
        configuration_directory_path: Path | None = None,
        *,
        locale: str | None = None,
    ):
        if configuration_directory_path is None:
            deprecate(
                f"Initializing {type(self)} without a configuration directory path is deprecated as of Betty 0.3.3, and will be removed in Betty 0.4.x.",
                stacklevel=2,
            )
            configuration_directory_path = CONFIGURATION_DIRECTORY_PATH
        super().__init__()
        self._configuration_directory_path = configuration_directory_path
        self._locale: str | None = locale

    @property
    def configuration_file_path(self) -> Path:
        return self._configuration_directory_path / "app.json"

    @configuration_file_path.setter
    def configuration_file_path(self, __) -> None:
        pass

    @configuration_file_path.deleter
    def configuration_file_path(self) -> None:
        pass

    @property
    def locale(self) -> str | None:
        return self._locale

    @locale.setter
    def locale(self, locale: str) -> None:
        try:
            get_data(locale)
        except ValueError:
            raise AssertionFailed(
                Str._(
                    '"{locale}" is not a valid IETF BCP 47 language tag.',
                    locale=locale,
                )
            )
        self._locale = locale
        self._dispatch_change()

    def update(self, other: Self) -> None:
        self._locale = other._locale
        self._dispatch_change()

    @classmethod
    def load(
        cls,
        dump: Dump,
        configuration: Self | None = None,
    ) -> Self:
        if configuration is None:
            configuration = cls()
        asserter = Asserter()
        asserter.assert_record(
            Fields(
                OptionalField(
                    "locale",
                    Assertions(asserter.assert_str())
                    | asserter.assert_setattr(configuration, "locale"),
                ),
            ),
        )(dump)
        return configuration

    def dump(self) -> VoidableDump:
        return minimize({"locale": void_none(self.locale)}, True)


class _BackwardsCompatiblePickledFileCache(PickledFileCache[Any], FileCache):
    """
    Provide a Backwards Compatible cache.

    .. deprecated:: 0.3.3
       This class is deprecated as of Betty 0.3.3, and will be removed in Betty 0.4.x.
    """

    @property
    def path(self) -> Path:
        return self._path


@final
class App(Configurable[AppConfiguration]):
    def __init__(
        self,
        configuration: AppConfiguration | None = None,
        project: Project | None = None,
        cache_directory_path: Path | None = None,
    ):
        super().__init__()
        self._started = False
        if configuration is None:
            deprecate(
                f"Initializing {type(self)} without `configuration` is deprecated as of Betty 0.3.2, and will be removed in Betty 0.4.x.",
                stacklevel=2,
            )
        if cache_directory_path is None:
            deprecate(
                f"Initializing {type(self)} without `cache_directory_path` is deprecated as of Betty 0.3.2, and will be removed in Betty 0.4.x.",
                stacklevel=2,
            )
        self._configuration = configuration or AppConfiguration()
        self._configuration.on_change(self._on_locale_change)
        self._assets: FileSystem | None = None
        self._extensions = _AppExtensions()
        self._extensions_initialized = False
        self._localization_initialized = False
        self._localizer: Localizer | None = None
        self._localizers: LocalizerRepository | None = None
        with suppress(FileNotFoundError):
            wait_to_thread(self.configuration.read())
        self._project = project or Project()
        self.project.configuration.extensions.on_change(self._update_extensions)

        self._dispatcher: ExtensionDispatcher | None = None
        self._entity_types: set[type[Entity]] | None = None
        self._event_types: set[type[EventType]] | None = None
        self._url_generator: LocalizedUrlGenerator | None = None
        self._static_url_generator: StaticUrlGenerator | None = None
        self._jinja2_environment: Environment | None = None
        self._renderer: Renderer | None = None
        self._http_client: aiohttp.ClientSession | None = None
        self._cache_directory_path = (
            CACHE_DIRECTORY_PATH
            if cache_directory_path is None
            else cache_directory_path
        )
        self._cache: Cache[Any] & FileCache | None = None
        self._binary_file_cache: BinaryFileCache | None = None
        self._process_pool: Executor | None = None

    @classmethod
    @asynccontextmanager
    async def new_from_environment(
        cls,
        *,
        project: Project | None = None,
    ) -> AsyncIterator[Self]:
        yield cls(
            AppConfiguration(CONFIGURATION_DIRECTORY_PATH),
            project,
            Path(environ.get("BETTY_CACHE_DIRECTORY", CACHE_DIRECTORY_PATH)),
        )

    @classmethod
    @asynccontextmanager
    async def new_from_app(
        cls,
        app: App,
        *,
        project: Project | None = None,
    ) -> AsyncIterator[Self]:
        yield cls(
            AppConfiguration(app.configuration._configuration_directory_path),
            app.project if project is None else project,
            app._cache_directory_path,
        )

    @classmethod
    @asynccontextmanager
    async def new_temporary(
        cls,
        *,
        project: Project | None = None,
    ) -> AsyncIterator[Self]:
        async with (
            TemporaryDirectory() as configuration_directory_path_str,
            TemporaryDirectory() as cache_directory_path_str,
        ):
            yield cls(
                AppConfiguration(Path(configuration_directory_path_str)),
                project,
                cache_directory_path=Path(cache_directory_path_str),
            )

    async def __aenter__(self) -> Self:
        await self.start()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.stop()

    async def start(self) -> None:
        if self._started:
            raise RuntimeError("This app has started already.")
        self._started = True

    async def stop(self) -> None:
        del self.http_client
        self._started = False

    def __del__(self) -> None:
        if self._started:
            raise RuntimeError(f"{self} was started, but never stopped.")

    def _on_locale_change(self) -> None:
        del self.localizer
        del self.localizers

    @property
    def project(self) -> Project:
        return self._project

    def discover_extension_types(self) -> set[type[Extension]]:
        from betty.app import extension

        return {
            *extension.discover_extension_types(),
            *map(type, self._extensions.flatten()),
        }

    @property
    def extensions(self) -> Extensions:
        if not self._extensions_initialized:
            self._extensions_initialized = True
            self._update_extensions()

        return self._extensions

    def _update_extensions(self) -> None:
        extension_types_enabled_in_configuration = set()
        for (
            app_extension_configuration
        ) in self.project.configuration.extensions.values():
            if app_extension_configuration.enabled:
                app_extension_configuration.extension_type.enable_requirement().assert_met()
                extension_types_enabled_in_configuration.add(
                    app_extension_configuration.extension_type
                )

        extension_types_sorter = TopologicalSorter(
            build_extension_type_graph(extension_types_enabled_in_configuration)
        )
        try:
            extension_types_sorter.prepare()
        except CycleError:
            raise CyclicDependencyError(
                [
                    app_extension_configuration.extension_type
                    for app_extension_configuration in self.project.configuration.extensions.values()
                ]
            )

        extensions = []
        while extension_types_sorter.is_active():
            extension_types_batch = extension_types_sorter.get_ready()
            extensions_batch = []
            for extension_type in extension_types_batch:
                if (
                    issubclass(extension_type, ConfigurableExtension)
                    and extension_type in self.project.configuration.extensions
                ):
                    extension: Extension = extension_type(
                        self,
                        configuration=self.project.configuration.extensions[
                            extension_type
                        ].extension_configuration,
                    )
                else:
                    extension = extension_type(self)
                extensions_batch.append(extension)
                extension_types_sorter.done(extension_type)
            extensions.append(
                sorted(extensions_batch, key=lambda extension: extension.name())
            )
        self._extensions._update(extensions)
        del self.assets
        del self.localizers
        del self.localizer
        del self.jinja2_environment
        del self.renderer
        del self.entity_types
        del self.event_types

    @property
    def assets(self) -> FileSystem:
        if self._assets is None:
            assets = FileSystem()
            assets.prepend(fs.ASSETS_DIRECTORY_PATH, "utf-8")
            for extension in self.extensions.flatten():
                extension_assets_directory_path = extension.assets_directory_path()
                if extension_assets_directory_path is not None:
                    assets.prepend(extension_assets_directory_path, "utf-8")
            assets.prepend(self.project.configuration.assets_directory_path)
            self._assets = assets
        return self._assets

    @assets.deleter
    def assets(self) -> None:
        self._assets = None

    @property
    def dispatcher(self) -> Dispatcher:
        if self._dispatcher is None:
            self._dispatcher = ExtensionDispatcher(self.extensions)

        return self._dispatcher

    @property
    def url_generator(self) -> LocalizedUrlGenerator:
        from betty.url import AppUrlGenerator

        if self._url_generator is None:
            self._url_generator = AppUrlGenerator(self)
        return self._url_generator

    @property
    def static_url_generator(self) -> StaticUrlGenerator:
        from betty.url import StaticPathUrlGenerator

        if self._static_url_generator is None:
            self._static_url_generator = StaticPathUrlGenerator(
                self.project.configuration
            )
        return self._static_url_generator

    @property
    def localizer(self) -> Localizer:
        """
        Get the application's localizer.
        """
        if self._localizer is None:
            self._localizer = wait_to_thread(
                self.localizers.get_negotiated(
                    self.configuration.locale or DEFAULT_LOCALE
                )
            )
        return self._localizer

    @localizer.deleter
    def localizer(self) -> None:
        self._localizer = None
        del self.cache
        del self.binary_file_cache

    @property
    def localizers(self) -> LocalizerRepository:
        if self._localizers is None:
            self._localizers = LocalizerRepository(self.assets)
        return self._localizers

    @localizers.deleter
    def localizers(self) -> None:
        self._localizers = None

    @property
    def jinja2_environment(self) -> Environment:
        if not self._jinja2_environment:
            from betty.jinja2 import Environment

            self._jinja2_environment = Environment(self)

        return self._jinja2_environment

    @jinja2_environment.deleter
    def jinja2_environment(self) -> None:
        self._jinja2_environment = None

    @property
    def renderer(self) -> Renderer:
        if not self._renderer:
            from betty.jinja2 import Jinja2Renderer

            self._renderer = SequentialRenderer(
                [
                    Jinja2Renderer(self.jinja2_environment, self.project.configuration),
                ]
            )

        return self._renderer

    @renderer.deleter
    def renderer(self) -> None:
        self._renderer = None

    @property
    def http_client(self) -> aiohttp.ClientSession:
        if not self._http_client:
            self._http_client = aiohttp.ClientSession(
                connector=aiohttp.TCPConnector(limit_per_host=5),
                headers={
                    "User-Agent": f"Betty (https://github.com/bartfeenstra/betty) on behalf of {self._project.configuration.base_url}{self._project.configuration.root_path}",
                },
            )
            weakref.finalize(
                self,
                lambda: (
                    None
                    if self._http_client is None
                    else wait_to_thread(self._http_client.close())
                ),
            )
        return self._http_client

    @http_client.deleter
    def http_client(self) -> None:
        if self._http_client is not None:
            wait_to_thread(self._http_client.close())
            self._http_client = None

    @property
    def entity_types(self) -> set[type[Entity]]:
        if self._entity_types is None:
            from betty.model.ancestry import (
                Citation,
                Enclosure,
                Event,
                File,
                Note,
                Person,
                PersonName,
                Presence,
                Place,
                Source,
            )

            self._entity_types = reduce(
                operator.or_,
                wait_to_thread(self.dispatcher.dispatch(EntityTypeProvider)()),
                set(),
            ) | {
                Citation,
                Enclosure,
                Event,
                File,
                Note,
                Person,
                PersonName,
                Presence,
                Place,
                Source,
            }
        return self._entity_types

    @entity_types.deleter
    def entity_types(self) -> None:
        self._entity_types = None

    @property
    def event_types(self) -> set[type[EventType]]:
        if self._event_types is None:
            self._event_types = set(
                wait_to_thread(self.dispatcher.dispatch(EventTypeProvider)())
            ) | {
                Birth,
                Baptism,
                Adoption,
                Death,
                Funeral,
                Cremation,
                Burial,
                Will,
                Engagement,
                Marriage,
                MarriageAnnouncement,
                Divorce,
                DivorceAnnouncement,
                Residence,
                Immigration,
                Emigration,
                Occupation,
                Retirement,
                Correspondence,
                Confirmation,
            }
        return self._event_types

    @event_types.deleter
    def event_types(self) -> None:
        self._event_types = None

    @property
    def servers(self) -> Mapping[str, Server]:
        from betty import serve
        from betty.extension.demo import DemoServer

        return {
            server.name(): server
            for server in [
                *(
                    server
                    for extension in self.extensions.flatten()
                    if isinstance(extension, serve.ServerProvider)
                    for server in extension.servers
                ),
                serve.BuiltinAppServer(self),
                DemoServer(app=self),
            ]
        }

    @property
    def cache(self) -> Cache[Any] & FileCache:
        if self._cache is None:
            self._cache = _BackwardsCompatiblePickledFileCache(
                self.localizer, self._cache_directory_path
            )
        return self._cache

    @cache.deleter
    def cache(self) -> None:
        self._cache = None

    @property
    def binary_file_cache(self) -> BinaryFileCache:
        if self._binary_file_cache is None:
            self._binary_file_cache = BinaryFileCache(
                self.localizer, self._cache_directory_path
            )
        return self._binary_file_cache

    @binary_file_cache.deleter
    def binary_file_cache(self) -> None:
        self._binary_file_cache = None

    @property
    def process_pool(self) -> Executor:
        if self._process_pool is None:
            # Avoid `fork` so as not to start worker processes with unneeded resources.
            # Settle for `spawn` so all environments use the same start method.
            self._process_pool = ProcessPoolExecutor(mp_context=get_context("spawn"))
        return self._process_pool



"""
Provide the project API.
"""

from __future__ import annotations

from contextlib import suppress
from pathlib import Path
from reprlib import recursive_repr
from typing import Any, Generic, final, Iterable, cast, Self
from urllib.parse import urlparse

from betty.app.extension import Extension, ConfigurableExtension
from betty.classtools import repr_instance
from betty.config import (
    Configuration,
    Configurable,
    FileBasedConfiguration,
    ConfigurationMapping,
    ConfigurationSequence,
)
from betty.hashid import hashid
from betty.locale import get_data, Str
from betty.model import Entity, get_entity_type_name, UserFacingEntity, EntityT
from betty.model.ancestry import Ancestry, Person, Event, Place, Source
from betty.serde.dump import (
    Dump,
    VoidableDump,
    void_none,
    minimize,
    Void,
    VoidableDictDump,
)
from betty.serde.load import (
    AssertionFailed,
    Fields,
    Assertions,
    Assertion,
    RequiredField,
    OptionalField,
    Asserter,
)
from betty.warnings import deprecate

DEFAULT_LIFETIME_THRESHOLD = 125


class EntityReference(Configuration, Generic[EntityT]):
    def __init__(
        self,
        entity_type: type[EntityT] | None = None,
        entity_id: str | None = None,
        *,
        entity_type_is_constrained: bool = False,
    ):
        super().__init__()
        self._entity_type = entity_type
        self._entity_id = entity_id
        self._entity_type_is_constrained = entity_type_is_constrained

    @property
    def entity_type(self) -> type[EntityT] | None:
        return self._entity_type

    @entity_type.setter
    def entity_type(self, entity_type: type[EntityT]) -> None:
        if self._entity_type_is_constrained:
            raise AttributeError(
                f"The entity type cannot be set, as it is already constrained to {self._entity_type}."
            )
        self._entity_type = entity_type
        self._dispatch_change()

    @property
    def entity_id(self) -> str | None:
        return self._entity_id

    @entity_id.setter
    def entity_id(self, entity_id: str) -> None:
        self._entity_id = entity_id
        self._dispatch_change()

    @entity_id.deleter
    def entity_id(self) -> None:
        self._entity_id = None

    @property
    def entity_type_is_constrained(self) -> bool:
        return self._entity_type_is_constrained

    def update(self, other: Self) -> None:
        self._entity_type = other._entity_type
        self._entity_type_is_constrained = other._entity_type_is_constrained
        self._entity_id = other._entity_id
        self._dispatch_change()

    @classmethod
    def load(
        cls,
        dump: Dump,
        configuration: Self | None = None,
    ) -> Self:
        if configuration is None:
            configuration = cls()
        asserter = Asserter()
        if isinstance(dump, dict) or not configuration.entity_type_is_constrained:
            asserter.assert_record(
                Fields(
                    RequiredField(
                        "entity_type",
                        Assertions(asserter.assert_entity_type())
                        | asserter.assert_setattr(configuration, "entity_type"),
                    ),
                    OptionalField(
                        "entity_id",
                        Assertions(asserter.assert_str())
                        | asserter.assert_setattr(configuration, "entity_id"),
                    ),
                )
            )(dump)
        else:
            asserter.assert_str()(dump)
            asserter.assert_setattr(configuration, "entity_id")(dump)  # type: ignore[arg-type]
        return configuration

    def dump(self) -> VoidableDump:
        if self.entity_type_is_constrained:
            return void_none(self.entity_id)

        if self.entity_type is None or self.entity_id is None:
            return Void

        dump: VoidableDictDump[VoidableDump] = {
            "entity_type": (
                get_entity_type_name(self._entity_type) if self._entity_type else Void
            ),
            "entity_id": self._entity_id,
        }

        return minimize(dump)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, EntityReference):
            return NotImplemented
        return (
            self.entity_type == other.entity_type and self.entity_id == other.entity_id
        )


class EntityReferenceSequence(
    Generic[EntityT], ConfigurationSequence[EntityReference[EntityT]]
):
    def __init__(
        self,
        entity_references: Iterable[EntityReference[EntityT]] | None = None,
        *,
        entity_type_constraint: type[EntityT] | None = None,
    ):
        self._entity_type_constraint = entity_type_constraint
        super().__init__(entity_references)

    @classmethod
    def _item_type(cls) -> type[EntityReference[EntityT]]:
        return EntityReference

    def _on_add(self, configuration: EntityReference[EntityT]) -> None:
        super()._on_add(configuration)

        entity_type_constraint = self._entity_type_constraint
        entity_reference_entity_type = configuration._entity_type

        if entity_type_constraint is None:
            return

        if (
            entity_reference_entity_type == entity_type_constraint
            and configuration.entity_type_is_constrained
        ):
            return

        expected_entity_type_name = get_entity_type_name(
            cast(type[Entity], entity_type_constraint),
        )
        expected_entity_type_label = entity_type_constraint.entity_type_label()

        if entity_reference_entity_type is None:
            raise AssertionFailed(
                Str._(
                    "The entity reference must be for an entity of type {expected_entity_type_name} ({expected_entity_type_label}), but instead does not specify an entity type at all.",
                    expected_entity_type_name=expected_entity_type_name,
                    expected_entity_type_label=expected_entity_type_label,
                )
            )

        actual_entity_type_label = entity_type_constraint.entity_type_label()

        raise AssertionFailed(
            Str._(
                "The entity reference must be for an entity of type {expected_entity_type_name} ({expected_entity_type_label}), but instead is for an entity of type {actual_entity_type_name} ({actual_entity_type_label})",
                expected_entity_type_name=expected_entity_type_name,
                expected_entity_type_label=expected_entity_type_label,
                actual_entity_type_name=get_entity_type_name(
                    entity_reference_entity_type
                ),
                actual_entity_type_label=actual_entity_type_label,
            )
        )


class ExtensionConfiguration(Configuration):
    def __init__(
        self,
        extension_type: type[Extension],
        *,
        enabled: bool = True,
        extension_configuration: Configuration | None = None,
    ):
        super().__init__()
        self._extension_type = extension_type
        self._enabled = enabled
        if extension_configuration is None and issubclass(
            extension_type, ConfigurableExtension
        ):
            extension_configuration = extension_type.default_configuration()
        if extension_configuration is not None:
            extension_configuration.on_change(self)
        self._extension_configuration = extension_configuration

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        if self.extension_type != other.extension_type:
            return False
        if self.enabled != other.enabled:
            return False
        if self.extension_configuration != other.extension_configuration:
            return False
        return True

    @property
    def extension_type(self) -> type[Extension]:
        return self._extension_type

    @property
    def enabled(self) -> bool:
        return self._enabled

    @enabled.setter
    def enabled(self, enabled: bool) -> None:
        self._enabled = enabled
        self._dispatch_change()

    @property
    def extension_configuration(self) -> Configuration | None:
        return self._extension_configuration

    def update(self, other: Self) -> None:
        self._extension_type = other._extension_type
        self._enabled = other._enabled
        self._extension_configuration = other._extension_configuration

    @classmethod
    def load(
        cls,
        dump: Dump,
        configuration: Self | None = None,
    ) -> Self:
        asserter = Asserter()
        extension_type = asserter.assert_field(
            RequiredField(
                "extension",
                Assertions(asserter.assert_extension_type()),
            )
        )(dump)
        if configuration is None:
            configuration = cls(extension_type)
        else:
            # This MUST NOT fail. If it does, this is a bug in the calling code that must be fixed.
            assert extension_type is configuration.extension_type
        asserter.assert_record(
            Fields(
                RequiredField(
                    "extension",
                ),
                OptionalField(
                    "enabled",
                    Assertions(asserter.assert_bool())
                    | asserter.assert_setattr(configuration, "enabled"),
                ),
                OptionalField(
                    "configuration",
                    Assertions(
                        configuration._assert_load_extension_configuration(
                            configuration.extension_type
                        )
                    ),
                ),
            )
        )(dump)
        return configuration

    def _assert_load_extension_configuration(
        self, extension_type: type[Extension]
    ) -> Assertion[Any, Configuration]:
        def _assertion(value: Any) -> Configuration:
            extension_configuration = self._extension_configuration
            if isinstance(extension_configuration, Configuration):
                return extension_configuration.load(value, extension_configuration)
            raise AssertionFailed(
                Str._(
                    "{extension_type} is not configurable.",
                    extension_type=extension_type.name(),
                )
            )

        return _assertion

    def dump(self) -> VoidableDump:
        return minimize(
            {
                "extension": self.extension_type.name(),
                "enabled": self.enabled,
                "configuration": (
                    minimize(self.extension_configuration.dump())
                    if issubclass(self.extension_type, Configurable)
                    and self.extension_configuration
                    else Void
                ),
            }
        )


class ExtensionConfigurationMapping(
    ConfigurationMapping[type[Extension], ExtensionConfiguration]
):
    def _minimize_item_dump(self) -> bool:
        return True

    @classmethod
    def _create_default_item(
        cls, configuration_key: type[Extension]
    ) -> ExtensionConfiguration:
        return ExtensionConfiguration(configuration_key)

    def __init__(
        self,
        configurations: Iterable[ExtensionConfiguration] | None = None,
    ):
        super().__init__(configurations)

    @classmethod
    def _item_type(cls) -> type[ExtensionConfiguration]:
        return ExtensionConfiguration

    def _get_key(self, configuration: ExtensionConfiguration) -> type[Extension]:
        return configuration.extension_type

    @classmethod
    def _load_key(
        cls,
        item_dump: Dump,
        key_dump: str,
    ) -> Dump:
        asserter = Asserter()
        dict_dump = asserter.assert_dict()(item_dump)
        dict_dump["extension"] = key_dump
        return dict_dump

    def _dump_key(self, item_dump: VoidableDump) -> tuple[VoidableDump, str]:
        dict_dump = self._asserter.assert_dict()(item_dump)
        return dict_dump, dict_dump.pop("extension")

    def enable(self, *extension_types: type[Extension]) -> None:
        for extension_type in extension_types:
            try:
                self._configurations[extension_type].enabled = True
            except KeyError:
                self.append(ExtensionConfiguration(extension_type))

    def disable(self, *extension_types: type[Extension]) -> None:
        for extension_type in extension_types:
            with suppress(KeyError):
                self._configurations[extension_type].enabled = False


class EntityTypeConfiguration(Configuration):
    def __init__(
        self,
        entity_type: type[Entity],
        *,
        generate_html_list: bool | None = None,
    ):
        super().__init__()
        self._entity_type = entity_type
        self.generate_html_list = generate_html_list  # type: ignore[assignment]

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        if self.entity_type != other.entity_type:
            return False
        if self.generate_html_list != other.generate_html_list:
            return False
        return True

    @property
    def entity_type(self) -> type[Entity]:
        return self._entity_type

    @property
    def generate_html_list(self) -> bool:
        return self._generate_html_list or False

    @generate_html_list.setter
    def generate_html_list(self, generate_html_list: bool | None) -> None:
        if generate_html_list and not issubclass(self._entity_type, UserFacingEntity):
            raise AssertionFailed(
                Str._(
                    "Cannot generate pages for {entity_type}, because it is not a user-facing entity type.",
                    entity_type=get_entity_type_name(self._entity_type),
                )
            )
        self._generate_html_list = generate_html_list
        self._dispatch_change()

    def update(self, other: Self) -> None:
        self._entity_type = other._entity_type
        self._generate_html_list = other._generate_html_list
        self._dispatch_change()

    @classmethod
    def load(
        cls,
        dump: Dump,
        configuration: Self | None = None,
    ) -> Self:
        asserter = Asserter()
        entity_type = asserter.assert_field(
            RequiredField[Any, type[Entity]](
                "entity_type",
                Assertions(asserter.assert_str()) | asserter.assert_entity_type(),
            ),
        )(dump)
        configuration = cls(entity_type)
        asserter.assert_record(
            Fields(
                OptionalField(
                    "entity_type",
                ),
                OptionalField(
                    "generate_html_list",
                    Assertions(asserter.assert_bool())
                    | asserter.assert_setattr(configuration, "generate_html_list"),
                ),
            )
        )(dump)
        return configuration

    def dump(self) -> VoidableDump:
        dump: VoidableDictDump[VoidableDump] = {
            "entity_type": get_entity_type_name(self._entity_type),
            "generate_html_list": (
                Void if self._generate_html_list is None else self._generate_html_list
            ),
        }

        return minimize(dump)


class EntityTypeConfigurationMapping(
    ConfigurationMapping[type[Entity], EntityTypeConfiguration]
):
    def _minimize_item_dump(self) -> bool:
        return True

    def _get_key(self, configuration: EntityTypeConfiguration) -> type[Entity]:
        return configuration.entity_type

    @classmethod
    def _load_key(
        cls,
        item_dump: Dump,
        key_dump: str,
    ) -> Dump:
        asserter = Asserter()
        dict_dump = asserter.assert_dict()(item_dump)
        asserter.assert_entity_type()(key_dump)
        dict_dump["entity_type"] = key_dump
        return dict_dump

    def _dump_key(self, item_dump: VoidableDump) -> tuple[VoidableDump, str]:
        dict_dump = self._asserter.assert_dict()(item_dump)
        return dict_dump, dict_dump.pop("entity_type")

    @classmethod
    def _item_type(cls) -> type[EntityTypeConfiguration]:
        return EntityTypeConfiguration

    @classmethod
    def _create_default_item(
        cls, configuration_key: type[Entity]
    ) -> EntityTypeConfiguration:
        return EntityTypeConfiguration(configuration_key)


class LocaleConfiguration(Configuration):
    def __init__(
        self,
        locale: str,
        *,
        alias: str | None = None,
    ):
        super().__init__()
        self._locale = locale
        if alias is not None and "/" in alias:
            raise AssertionFailed(Str._("Locale aliases must not contain slashes."))
        self._alias = alias

    @recursive_repr()
    def __repr__(self) -> str:
        return repr_instance(self, locale=self.locale, alias=self.alias)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        if self.locale != other.locale:
            return False
        if self.alias != other.alias:
            return False
        return True

    def __hash__(self) -> int:
        return hash((self._locale, self._alias))

    @property
    def locale(self) -> str:
        return self._locale

    @property
    def alias(self) -> str:
        if self._alias is None:
            return self.locale
        return self._alias

    @alias.setter
    def alias(self, alias: str | None) -> None:
        self._alias = alias
        self._dispatch_change()

    def update(self, other: Self) -> None:
        self._locale = other._locale
        self._alias = other._alias

    @classmethod
    def load(
        cls,
        dump: Dump,
        configuration: Self | None = None,
    ) -> Self:
        asserter = Asserter()
        locale = asserter.assert_field(
            RequiredField("locale", Assertions(asserter.assert_locale())),
        )(dump)
        if configuration is None:
            configuration = cls(locale)
        asserter.assert_record(
            Fields(
                RequiredField("locale"),
                OptionalField(
                    "alias",
                    Assertions(asserter.assert_str())
                    | asserter.assert_setattr(configuration, "alias"),
                ),
            )
        )(dump)
        return configuration

    def dump(self) -> VoidableDump:
        return minimize({"locale": self.locale, "alias": void_none(self._alias)})


class LocaleConfigurationMapping(ConfigurationMapping[str, LocaleConfiguration]):
    @classmethod
    def _create_default_item(cls, configuration_key: str) -> LocaleConfiguration:
        return LocaleConfiguration(configuration_key)

    def __init__(
        self,
        configurations: Iterable[LocaleConfiguration] | None = None,
    ):
        super().__init__(configurations)
        if len(self) == 0:
            self.append(LocaleConfiguration("en-US"))

    def _get_key(self, configuration: LocaleConfiguration) -> str:
        return configuration.locale

    @classmethod
    def _load_key(
        cls,
        item_dump: Dump,
        key_dump: str,
    ) -> Dump:
        asserter = Asserter()
        dict_item_dump = asserter.assert_dict()(item_dump)
        dict_item_dump["locale"] = key_dump
        return dict_item_dump

    def _dump_key(self, item_dump: VoidableDump) -> tuple[VoidableDump, str]:
        dict_item_dump = self._asserter.assert_dict()(item_dump)
        return dict_item_dump, dict_item_dump.pop("locale")

    @classmethod
    def _item_type(cls) -> type[LocaleConfiguration]:
        return LocaleConfiguration

    def _on_remove(self, configuration: LocaleConfiguration) -> None:
        if len(self._configurations) <= 1:
            raise AssertionFailed(
                Str._(
                    "Cannot remove the last remaining locale {locale}",
                    locale=get_data(configuration.locale).get_display_name()
                    or configuration.locale,
                )
            )

    @property
    def default(self) -> LocaleConfiguration:
        return next(iter(self._configurations.values()))

    @default.setter
    def default(self, configuration: LocaleConfiguration | str) -> None:
        if isinstance(configuration, str):
            configuration = self[configuration]
        self._configurations[configuration.locale] = configuration
        self._configurations.move_to_end(configuration.locale, False)
        self._dispatch_change()

    @property
    def multilingual(self) -> bool:
        return len(self) > 1


@final
class ProjectConfiguration(FileBasedConfiguration):
    def __init__(
        self,
        base_url: str | None = None,
        root_path: str = "",
        clean_urls: bool = False,
        title: str = "Betty",
        author: str | None = None,
        entity_types: Iterable[EntityTypeConfiguration] | None = None,
        extensions: Iterable[ExtensionConfiguration] | None = None,
        debug: bool = False,
        locales: Iterable[LocaleConfiguration] | None = None,
        lifetime_threshold: int = DEFAULT_LIFETIME_THRESHOLD,
        name: str | None = None,
    ):
        super().__init__()
        self._name = name
        self._computed_name: str | None = None
        self._base_url = "https://example.com" if base_url is None else base_url
        self._root_path = root_path
        self._clean_urls = clean_urls
        self._title = title
        self._author = author
        self._entity_types = EntityTypeConfigurationMapping(
            entity_types
            or [
                EntityTypeConfiguration(
                    entity_type=Person,
                    generate_html_list=True,
                ),
                EntityTypeConfiguration(
                    entity_type=Event,
                    generate_html_list=True,
                ),
                EntityTypeConfiguration(
                    entity_type=Place,
                    generate_html_list=True,
                ),
                EntityTypeConfiguration(
                    entity_type=Source,
                    generate_html_list=True,
                ),
            ]
        )
        self._entity_types.on_change(self)
        self._extensions = ExtensionConfigurationMapping(extensions or ())
        self._extensions.on_change(self)
        self._debug = debug
        self._locales = LocaleConfigurationMapping(locales or ())
        self._locales.on_change(self)
        self._lifetime_threshold = lifetime_threshold

    @property
    def name(self) -> str | None:
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        self._name = name
        self._dispatch_change()

    @property
    def project_directory_path(self) -> Path:
        return self.configuration_file_path.parent

    @property
    def output_directory_path(self) -> Path:
        return self.project_directory_path / "output"

    @property
    def assets_directory_path(self) -> Path:
        return self.project_directory_path / "assets"

    @property
    def www_directory_path(self) -> Path:
        return self.output_directory_path / "www"

    def localize_www_directory_path(self, locale: str) -> Path:
        if self.locales.multilingual:
            return self.www_directory_path / self.locales[locale].alias
        return self.www_directory_path

    @property
    def title(self) -> str:
        return self._title

    @title.setter
    def title(self, title: str) -> None:
        self._title = title
        self._dispatch_change()

    @property
    def author(self) -> str | None:
        return self._author

    @author.setter
    def author(self, author: str | None) -> None:
        self._author = author
        self._dispatch_change()

    @property
    def base_url(self) -> str:
        return self._base_url

    @base_url.setter
    def base_url(self, base_url: str) -> None:
        base_url_parts = urlparse(base_url)
        if not base_url_parts.scheme:
            raise AssertionFailed(
                Str._(
                    "The base URL must start with a scheme such as https://, http://, or file://."
                )
            )
        if not base_url_parts.netloc:
            raise AssertionFailed(Str._("The base URL must include a path."))
        self._base_url = "%s://%s" % (base_url_parts.scheme, base_url_parts.netloc)
        self._dispatch_change()

    @property
    def root_path(self) -> str:
        return self._root_path

    @root_path.setter
    def root_path(self, root_path: str) -> None:
        self._root_path = root_path.strip("/")
        self._dispatch_change()

    @property
    def clean_urls(self) -> bool:
        return self._clean_urls

    @clean_urls.setter
    def clean_urls(self, clean_urls: bool) -> None:
        self._clean_urls = clean_urls
        self._dispatch_change()

    @property
    def locales(self) -> LocaleConfigurationMapping:
        return self._locales

    @property
    def entity_types(self) -> EntityTypeConfigurationMapping:
        return self._entity_types

    @property
    def extensions(self) -> ExtensionConfigurationMapping:
        return self._extensions

    @property
    def debug(self) -> bool:
        return self._debug

    @debug.setter
    def debug(self, debug: bool) -> None:
        self._debug = debug
        self._dispatch_change()

    @property
    def lifetime_threshold(self) -> int:
        return self._lifetime_threshold

    @lifetime_threshold.setter
    def lifetime_threshold(self, lifetime_threshold: int) -> None:
        self._asserter.assert_positive_number()(lifetime_threshold)
        self._lifetime_threshold = lifetime_threshold
        self._dispatch_change()

    def update(self, other: Self) -> None:
        self._base_url = other._base_url
        self._title = other._title
        self._author = other._author
        self._root_path = other._root_path
        self._clean_urls = other._clean_urls
        self._debug = other._debug
        self._lifetime_threshold = other._lifetime_threshold
        self._locales.update(other._locales)
        self._extensions.update(other._extensions)
        self._entity_types.update(other._entity_types)
        self._dispatch_change()

    @classmethod
    def load(
        cls,
        dump: Dump,
        configuration: Self | None = None,
    ) -> Self:
        if configuration is None:
            configuration = cls()
        asserter = Asserter()
        asserter.assert_record(
            Fields(
                OptionalField(
                    "name",
                    Assertions(asserter.assert_str())
                    | asserter.assert_setattr(configuration, "name"),
                ),
                RequiredField(
                    "base_url",
                    Assertions(asserter.assert_str())
                    | asserter.assert_setattr(configuration, "base_url"),
                ),
                OptionalField(
                    "title",
                    Assertions(asserter.assert_str())
                    | asserter.assert_setattr(configuration, "title"),
                ),
                OptionalField(
                    "author",
                    Assertions(asserter.assert_str())
                    | asserter.assert_setattr(configuration, "author"),
                ),
                OptionalField(
                    "root_path",
                    Assertions(asserter.assert_str())
                    | asserter.assert_setattr(configuration, "root_path"),
                ),
                OptionalField(
                    "clean_urls",
                    Assertions(asserter.assert_bool())
                    | asserter.assert_setattr(configuration, "clean_urls"),
                ),
                OptionalField(
                    "debug",
                    Assertions(asserter.assert_bool())
                    | asserter.assert_setattr(configuration, "debug"),
                ),
                OptionalField(
                    "lifetime_threshold",
                    Assertions(asserter.assert_int())
                    | asserter.assert_setattr(configuration, "lifetime_threshold"),
                ),
                OptionalField(
                    "locales",
                    Assertions(
                        configuration._locales.assert_load(configuration.locales)
                    ),
                ),
                OptionalField(
                    "extensions",
                    Assertions(
                        configuration._extensions.assert_load(configuration.extensions)
                    ),
                ),
                OptionalField(
                    "entity_types",
                    Assertions(
                        configuration._entity_types.assert_load(
                            configuration.entity_types
                        )
                    ),
                ),
            )
        )(dump)
        return configuration

    def dump(self) -> VoidableDictDump[Dump]:
        return minimize(
            {  # type: ignore[return-value]
                "name": void_none(self.name),
                "base_url": self.base_url,
                "title": self.title,
                "root_path": void_none(self.root_path),
                "clean_urls": void_none(self.clean_urls),
                "author": void_none(self.author),
                "debug": void_none(self.debug),
                "lifetime_threshold": void_none(self.lifetime_threshold),
                "locales": self.locales.dump(),
                "extensions": self.extensions.dump(),
                "entity_types": self.entity_types.dump(),
            },
            True,
        )


class Project(Configurable[ProjectConfiguration]):
    def __init__(
        self,
        *,
        project_id: str | None = None,
        ancestry: Ancestry | None = None,
    ):
        super().__init__()
        if project_id is not None:
            deprecate(
                f"Initializing {type(self)} with a project ID is deprecated as of Betty 0.3.2, and will be removed in Betty 0.4.x. Instead, set {type(self)}.configuration.name.",
                stacklevel=2,
            )
        self._id = project_id
        self._configuration = ProjectConfiguration()
        self._ancestry = Ancestry() if ancestry is None else ancestry

    @property
    def id(self) -> str:
        deprecate(
            f"{type(self)}.id is deprecated as of Betty 0.3.2, and will be removed in Betty 0.4.x. Insead, use {type(self)}.name."
        )
        if self._id is None:
            return self.name
        return self._id

    @property
    def name(self) -> str:
        if self._configuration.name is None:
            return hashid(str(self._configuration.configuration_file_path))
        return self._configuration.name

    @property
    def ancestry(self) -> Ancestry:
        return self._ancestry



"""
Provide rendering utilities using `Jinja2 <https://jinja.palletsprojects.com>`_.
"""

from __future__ import annotations

import datetime
from collections import defaultdict
from collections.abc import MutableMapping, Iterator, Sequence
from pathlib import Path
from threading import Lock
from typing import Callable, Any, cast, TypeVar

import aiofiles
from aiofiles import os as aiofiles_os
from jinja2 import (
    Environment as Jinja2Environment,
    select_autoescape,
    FileSystemLoader,
    pass_context,
)
from jinja2.runtime import StrictUndefined, Context, DebugUndefined

from betty.app import App
from betty.app.extension import Extension
from betty.html import CssProvider, JsProvider
from betty.jinja2.filter import FILTERS
from betty.jinja2.test import TESTS
from betty.job import Context as JobContext
from betty.locale import Date, Localizer, DEFAULT_LOCALIZER
from betty.model import Entity, get_entity_type
from betty.model.ancestry import Citation
from betty.project import ProjectConfiguration
from betty.render import Renderer
from betty.serde.dump import Dumpable, DictDump, VoidableDump, Void, Dump

T = TypeVar("T")


def context_app(context: Context) -> App:
    """
    Get the current app from the Jinja2 context.
    """
    return cast(Environment, context.environment).app


def context_job_context(context: Context) -> JobContext | None:
    """
    Get the current job context from the Jinja2 context.
    """
    job_context = context.resolve_or_missing("job_context")
    return job_context if isinstance(job_context, JobContext) else None


def context_localizer(context: Context) -> Localizer:
    """
    Get the current localizer from the Jinja2 context.
    """
    localizer = context.resolve_or_missing("localizer")
    if isinstance(localizer, Localizer):
        return localizer
    raise RuntimeError(
        "No `localizer` context variable exists in this Jinja2 template."
    )


class _Citer:
    __slots__ = "_lock", "_cited"

    def __init__(self):
        self._lock = Lock()
        self._cited: list[Citation] = []

    def __iter__(self) -> enumerate[Citation]:
        return enumerate(self._cited, 1)

    def __len__(self) -> int:
        return len(self._cited)

    def cite(self, citation: Citation) -> int:
        with self._lock:
            if citation not in self._cited:
                self._cited.append(citation)
            return self._cited.index(citation) + 1


class _Breadcrumb(Dumpable):
    def __init__(self, label: str, url: str):
        self._label = label
        self._url = url

    def dump(self) -> DictDump[Dump]:
        return {
            "@type": "ListItem",
            "name": self._label,
            "item": self._url,
        }


class _Breadcrumbs(Dumpable):
    def __init__(self):
        self._breadcrumbs: list[_Breadcrumb] = []

    def append(self, label: str, url: str) -> None:
        self._breadcrumbs.append(_Breadcrumb(label, url))

    def dump(self) -> VoidableDump:
        if not self._breadcrumbs:
            return Void
        return {
            "@context": "https://schema.org",
            "@type": "BreadcrumbList",
            "itemListElement": [
                {
                    "position": position,
                    **breadcrumb.dump(),
                }
                for position, breadcrumb in enumerate(self._breadcrumbs, 1)
            ],
        }


class EntityContexts:
    def __init__(self, *entities: Entity) -> None:
        self._contexts: dict[type[Entity], Entity | None] = defaultdict(lambda: None)
        for entity in entities:
            self._contexts[entity.type] = entity

    def __getitem__(
        self, entity_type_or_type_name: type[Entity] | str
    ) -> Entity | None:
        if isinstance(entity_type_or_type_name, str):
            entity_type = get_entity_type(entity_type_or_type_name)
        else:
            entity_type = entity_type_or_type_name
        return self._contexts[entity_type]

    def __call__(self, *entities: Entity) -> EntityContexts:
        updated_contexts = EntityContexts()
        for entity in entities:
            updated_contexts._contexts[entity.type] = entity
        return updated_contexts


class Jinja2Provider:
    @property
    def globals(self) -> dict[str, Any]:
        return {}

    @property
    def filters(self) -> dict[str, Callable[..., Any]]:
        return {}

    @property
    def tests(self) -> dict[str, Callable[..., bool]]:
        return {}

    def new_context_vars(self) -> dict[str, Any]:
        return {}


class Environment(Jinja2Environment):
    globals: dict[str, Any]
    filters: dict[str, Callable[..., Any]]
    tests: dict[str, Callable[..., bool]]

    def __init__(self, app: App):
        template_directory_paths = [
            str(path / "templates") for path, _ in app.assets.paths
        ]
        super().__init__(
            loader=FileSystemLoader(template_directory_paths),
            auto_reload=app.project.configuration.debug,
            enable_async=True,
            undefined=(
                DebugUndefined if app.project.configuration.debug else StrictUndefined
            ),
            autoescape=select_autoescape(["html.j2"]),
            trim_blocks=True,
            lstrip_blocks=True,
            extensions=[
                "jinja2.ext.do",
                "jinja2.ext.i18n",
            ],
        )

        self._context_class: type[Context] | None = None
        self.app = app

        if app.project.configuration.debug:
            self.add_extension("jinja2.ext.debug")

        self._init_i18n()
        self._init_globals()
        self.filters.update(FILTERS)
        self.tests.update(TESTS)
        self._init_extensions()

    def _init_i18n(self) -> None:
        self.install_gettext_callables(  # type: ignore[attr-defined]
            gettext=self._gettext,
            ngettext=self._ngettext,
            pgettext=self._pgettext,
            npgettext=self._npgettext,
        )
        self.policies["ext.i18n.trimmed"] = True

    @property
    def context_class(self) -> type[Context]:  # type: ignore[override]
        if self._context_class is None:
            jinja2_providers: Sequence[Jinja2Provider & Extension] = [
                extension
                for extension in self.app.extensions.flatten()
                if isinstance(extension, Jinja2Provider)
            ]

            class _Context(Context):
                def __init__(
                    self,
                    environment: Environment,
                    parent: dict[str, Any],
                    name: str | None,
                    blocks: dict[str, Callable[[Context], Iterator[str]]],
                    globals: MutableMapping[str, Any] | None = None,
                ):
                    if "citer" not in parent:
                        parent["citer"] = _Citer()
                    if "breadcrumbs" not in parent:
                        parent["breadcrumbs"] = _Breadcrumbs()
                    for jinja2_provider in jinja2_providers:
                        for key, value in jinja2_provider.new_context_vars().items():
                            if key not in parent:
                                parent[key] = value
                    super().__init__(
                        environment,
                        parent,
                        name,
                        blocks,
                        globals,
                    )

            self._context_class = _Context

        return self._context_class

    @pass_context
    def _gettext(self, context: Context, message: str) -> str:
        return context_localizer(context).gettext(message)

    @pass_context
    def _ngettext(
        self, context: Context, message_singular: str, message_plural: str, n: int
    ) -> str:
        return context_localizer(context).ngettext(message_singular, message_plural, n)

    @pass_context
    def _pgettext(self, context: Context, gettext_context: str, message: str) -> str:
        return context_localizer(context).pgettext(gettext_context, message)

    @pass_context
    def _npgettext(
        self,
        context: Context,
        gettext_context: str,
        message_singular: str,
        message_plural: str,
        n: int,
    ) -> str:
        return context_localizer(context).npgettext(
            gettext_context, message_singular, message_plural, n
        )

    def _init_globals(self) -> None:
        self.globals["app"] = self.app
        today = datetime.date.today()
        self.globals["today"] = Date(today.year, today.month, today.day)
        # Ideally we would use the Dispatcher for this. However, it is asynchronous only.
        self.globals["public_css_paths"] = [
            path
            for extension in self.app.extensions.flatten()
            if isinstance(extension, CssProvider)
            for path in extension.public_css_paths
        ]
        self.globals["public_js_paths"] = [
            path
            for extension in self.app.extensions.flatten()
            if isinstance(extension, JsProvider)
            for path in extension.public_js_paths
        ]
        self.globals["entity_contexts"] = EntityContexts()
        self.globals["localizer"] = DEFAULT_LOCALIZER

    def _init_extensions(self) -> None:
        for extension in self.app.extensions.flatten():
            if isinstance(extension, Jinja2Provider):
                self.globals.update(extension.globals)
                self.filters.update(extension.filters)
                self.tests.update(extension.tests)


class Jinja2Renderer(Renderer):
    def __init__(self, environment: Environment, configuration: ProjectConfiguration):
        self._environment = environment
        self._configuration = configuration

    @property
    def file_extensions(self) -> set[str]:
        return {".j2"}

    async def render_file(
        self,
        file_path: Path,
        *,
        job_context: JobContext | None = None,
        localizer: Localizer | None = None,
    ) -> Path:
        destination_file_path = file_path.parent / file_path.stem
        data: dict[str, Any] = {}
        if job_context is not None:
            data["job_context"] = job_context
        if localizer is not None:
            data["localizer"] = localizer
        try:
            relative_file_destination_path = destination_file_path.relative_to(
                self._configuration.www_directory_path
            )
        except ValueError:
            pass
        else:
            resource = "/".join(relative_file_destination_path.parts)
            if self._configuration.locales.multilingual:
                resource_parts = resource.lstrip("/").split("/")
                if resource_parts[0] in map(
                    lambda x: x.alias, self._configuration.locales.values()
                ):
                    resource = "/".join(resource_parts[1:])
            data["page_resource"] = resource
        async with aiofiles.open(file_path) as f:
            template_source = await f.read()
        rendered = await self._environment.from_string(
            template_source, self._environment.globals
        ).render_async(data)
        async with aiofiles.open(destination_file_path, "w", encoding="utf-8") as f:
            await f.write(rendered)
        await aiofiles_os.remove(file_path)
        return destination_file_path



"""Provide the Graphical User Interface (GUI) for Betty Desktop."""

from __future__ import annotations

import pickle
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from typing import Any, TypeVar, Self

from PyQt6.QtCore import pyqtSlot, QObject, QCoreApplication
from PyQt6.QtGui import QPalette
from PyQt6.QtWidgets import QApplication, QWidget

from betty.app import App
from betty.gui.error import ExceptionError, _UnexpectedExceptionError
from betty.locale import Str
from betty.serde.format import FormatRepository

QWidgetT = TypeVar("QWidgetT", bound=QWidget)


def get_configuration_file_filter() -> Str:
    """
    Get the Qt file filter for project configuration files.
    """
    formats = FormatRepository()
    return Str._(
        "Betty project configuration ({supported_formats})",
        supported_formats=" ".join(
            f"*.{extension}"
            for format in formats.formats
            for extension in format.extensions
        ),
    )


class GuiBuilder:
    def gui_build(self) -> QWidget:
        raise NotImplementedError(repr(self))


def mark_valid(widget: QWidget) -> None:
    """
    Mark a widget as currently containing valid input.
    """
    widget.setProperty("invalid", "false")
    widget.setStyle(widget.style())
    widget.setToolTip("")


def mark_invalid(widget: QWidget, reason: str) -> None:
    """
    Mark a widget as currently containing invalid input.
    """
    widget.setProperty("invalid", "true")
    widget.setStyle(widget.style())
    widget.setToolTip(reason)


class BettyApplication(QApplication):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._app: App | None = None
        self.setApplicationName("Betty")
        self.setStyleSheet(self._stylesheet())

    def _is_dark_mode(self) -> bool:
        palette = self.palette()
        window_lightness = palette.color(QPalette.ColorRole.Window).lightness()
        window_text_lightness = palette.color(QPalette.ColorRole.WindowText).lightness()
        return window_lightness < window_text_lightness

    def _stylesheet(self) -> str:
        if self._is_dark_mode():
            caption_color = "#eeeeee"
        else:
            caption_color = "#333333"
        return f"""
            Caption {{
                color: {caption_color};
                font-size: 14px;
                margin-bottom: 0.3em;
            }}

            Code {{
                font-family: monospace;
            }}

            QLineEdit[invalid="true"] {{
                border: 1px solid red;
                color: red;
            }}

            QPushButton[pane-selector="true"] {{
                padding: 10px;
            }}

            LogRecord[level="50"],
            LogRecord[level="40"] {{
                color: red;
            }}

            LogRecord[level="30"] {{
                color: yellow;
            }}

            LogRecord[level="20"] {{
                color: green;
            }}

            LogRecord[level="10"],
            LogRecord[level="0"] {{
                color: white;
            }}

            _WelcomeText {{
                padding: 10px;
            }}

            _WelcomeTitle {{
                font-size: 20px;
                padding: 10px;
            }}

            _WelcomeHeading {{
                font-size: 16px;
                margin-top: 50px;
            }}

            _WelcomeAction {{
                padding: 10px;
            }}
            """

    @pyqtSlot(
        type,
        bytes,
        QObject,
        bool,
    )
    def _show_user_facing_error(
        self,
        error_type: type[Exception],
        pickled_error_message: bytes,
        parent: QObject,
        close_parent: bool,
    ) -> None:
        error_message = pickle.loads(pickled_error_message)
        window = ExceptionError(
            self.app,
            error_message,
            error_type,
            parent=parent,
            close_parent=close_parent,
        )
        window.show()

    @pyqtSlot(
        type,
        str,
        str,
        QObject,
        bool,
    )
    def _show_unexpected_exception(
        self,
        error_type: type[Exception],
        error_message: str,
        error_traceback: str,
        parent: QObject,
        close_parent: bool,
    ) -> None:
        window = _UnexpectedExceptionError(
            self.app,
            error_type,
            error_message,
            error_traceback,
            parent=parent,
            close_parent=close_parent,
        )
        window.show()

    @classmethod
    def instance(cls) -> Self:
        qapp = QCoreApplication.instance()
        assert isinstance(qapp, cls)
        return qapp

    @asynccontextmanager
    async def with_app(self, app: App) -> AsyncIterator[Self]:
        if self._app is not None:
            raise RuntimeError(f"This {type(self)} already has an {App}.")
        self._app = app
        yield self
        self._app = None

    @property
    def app(self) -> App:
        if self._app is None:
            raise RuntimeError(f"This {type(self)} does not have an {App} yet.")
        return self._app



"""
Provide Jinja2 filters to integrate with Webpack.
"""

from __future__ import annotations

from jinja2 import pass_context
from jinja2.runtime import Context

from betty.extension.webpack.jinja2 import _context_js_entrypoints
from betty.jinja2.filter import filter_public_js


@pass_context
def filter_webpack_entrypoint_js(context: Context, entrypoint_name: str) -> None:
    """
    Add a Webpack entrypoint's JavaScript files to the current page.
    """
    filter_public_js(context, "/js/webpack-entry-loader.js")
    _context_js_entrypoints(context).add(entrypoint_name)


FILTERS = {
    "webpack_entrypoint_js": filter_webpack_entrypoint_js,
}



"""
Provide string handling utilities.
"""

import re

_CAMEL_CASE_PATTERN = re.compile(r"(?<!^)(?=[A-Z])")


def camel_case_to_snake_case(string: str) -> str:
    """
    Convert camel case to snake case.
    """
    return _CAMEL_CASE_PATTERN.sub("_", string).lower()


def camel_case_to_kebab_case(string: str) -> str:
    """
    Convert camel case to kebab case.
    """
    return _CAMEL_CASE_PATTERN.sub("-", string).lower()


def upper_camel_case_to_lower_camel_case(string: str) -> str:
    """
    Convert upper camel case to lower camel case.
    """
    return string[0].lower() + string[1:]



"""
Provide Betty's ancestry event types.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from betty.locale import Str, DEFAULT_LOCALIZER

if TYPE_CHECKING:
    from betty.model.ancestry import Person


class EventTypeProvider:
    @property
    def entity_types(self) -> set[type[EventType]]:
        raise NotImplementedError(repr(self))


class EventType:
    def __new__(cls):
        raise RuntimeError("Event types cannot be instantiated.")

    @classmethod
    def name(cls) -> str:
        raise NotImplementedError(repr(cls))

    @classmethod
    def label(cls) -> Str:
        raise NotImplementedError(repr(cls))

    @classmethod
    def comes_before(cls) -> set[type[EventType]]:
        return set()

    @classmethod
    def comes_after(cls) -> set[type[EventType]]:
        return set()


class UnknownEventType(EventType):
    @classmethod
    def name(cls) -> str:
        return "unknown"

    @classmethod
    def label(cls) -> Str:
        return Str._("Unknown")


class DerivableEventType(EventType):
    pass


class CreatableDerivableEventType(DerivableEventType):
    @classmethod
    def may_create(cls, person: Person, lifetime_threshold: int) -> bool:
        return True


class PreBirthEventType(EventType):
    @classmethod
    def comes_before(cls) -> set[type[EventType]]:
        return {Birth}


class StartOfLifeEventType(EventType):
    pass


class DuringLifeEventType(EventType):
    @classmethod
    def comes_after(cls) -> set[type[EventType]]:
        return {Birth}

    @classmethod
    def comes_before(cls) -> set[type[EventType]]:
        return {Death}


class EndOfLifeEventType(EventType):
    pass


class PostDeathEventType(EventType):
    @classmethod
    def comes_after(cls) -> set[type[EventType]]:
        return {Death}


class Birth(CreatableDerivableEventType, StartOfLifeEventType):
    @classmethod
    def name(cls) -> str:
        return "birth"

    @classmethod
    def label(cls) -> Str:
        return Str._("Birth")

    @classmethod
    def comes_before(cls) -> set[type[EventType]]:
        return {DuringLifeEventType}


class Baptism(DuringLifeEventType, StartOfLifeEventType):
    @classmethod
    def name(cls) -> str:
        return "baptism"

    @classmethod
    def label(cls) -> Str:
        return Str._("Baptism")


class Adoption(DuringLifeEventType):
    @classmethod
    def name(cls) -> str:
        return "adoption"

    @classmethod
    def label(cls) -> Str:
        return Str._("Adoption")


class Death(CreatableDerivableEventType, EndOfLifeEventType):
    @classmethod
    def name(cls) -> str:
        return "death"

    @classmethod
    def label(cls) -> Str:
        return Str._("Death")

    @classmethod
    def comes_after(cls) -> set[type[EventType]]:
        return {DuringLifeEventType}

    @classmethod
    def may_create(cls, person: Person, lifetime_threshold: int) -> bool:
        from betty.privatizer import Privatizer

        return Privatizer(lifetime_threshold, localizer=DEFAULT_LOCALIZER).has_expired(
            person, 1
        )


class FinalDispositionEventType(
    PostDeathEventType, DerivableEventType, EndOfLifeEventType
):
    pass


class Funeral(FinalDispositionEventType):
    @classmethod
    def name(cls) -> str:
        return "funeral"

    @classmethod
    def label(cls) -> Str:
        return Str._("Funeral")


class Cremation(FinalDispositionEventType):
    @classmethod
    def name(cls) -> str:
        return "cremation"

    @classmethod
    def label(cls) -> Str:
        return Str._("Cremation")


class Burial(FinalDispositionEventType):
    @classmethod
    def name(cls) -> str:
        return "burial"

    @classmethod
    def label(cls) -> Str:
        return Str._("Burial")


class Will(PostDeathEventType):
    @classmethod
    def name(cls) -> str:
        return "will"

    @classmethod
    def label(cls) -> Str:
        return Str._("Will")


class Engagement(DuringLifeEventType):
    @classmethod
    def name(cls) -> str:
        return "engagement"

    @classmethod
    def label(cls) -> Str:
        return Str._("Engagement")

    @classmethod
    def comes_before(cls) -> set[type[EventType]]:
        return {Marriage}


class Marriage(DuringLifeEventType):
    @classmethod
    def name(cls) -> str:
        return "marriage"

    @classmethod
    def label(cls) -> Str:
        return Str._("Marriage")


class MarriageAnnouncement(DuringLifeEventType):
    @classmethod
    def name(cls) -> str:
        return "marriage-announcement"

    @classmethod
    def label(cls) -> Str:
        return Str._("Announcement of marriage")

    @classmethod
    def comes_before(cls) -> set[type[EventType]]:
        return {Marriage}


class Divorce(DuringLifeEventType):
    @classmethod
    def name(cls) -> str:
        return "divorce"

    @classmethod
    def label(cls) -> Str:
        return Str._("Divorce")

    @classmethod
    def comes_after(cls) -> set[type[EventType]]:
        return {Marriage}


class DivorceAnnouncement(DuringLifeEventType):
    @classmethod
    def name(cls) -> str:
        return "divorce-announcement"

    @classmethod
    def label(cls) -> Str:
        return Str._("Announcement of divorce")

    @classmethod
    def comes_after(cls) -> set[type[EventType]]:
        return {Marriage}

    @classmethod
    def comes_before(cls) -> set[type[EventType]]:
        return {Divorce}


class Residence(DuringLifeEventType):
    @classmethod
    def name(cls) -> str:
        return "residence"

    @classmethod
    def label(cls) -> Str:
        return Str._("Residence")


class Immigration(DuringLifeEventType):
    @classmethod
    def name(cls) -> str:
        return "immigration"

    @classmethod
    def label(cls) -> Str:
        return Str._("Immigration")


class Emigration(DuringLifeEventType):
    @classmethod
    def name(cls) -> str:
        return "emigration"

    @classmethod
    def label(cls) -> Str:
        return Str._("Emigration")


class Occupation(DuringLifeEventType):
    @classmethod
    def name(cls) -> str:
        return "occupation"

    @classmethod
    def label(cls) -> Str:
        return Str._("Occupation")


class Retirement(DuringLifeEventType):
    @classmethod
    def name(cls) -> str:
        return "retirement"

    @classmethod
    def label(cls) -> Str:
        return Str._("Retirement")


class Correspondence(EventType):
    @classmethod
    def name(cls) -> str:
        return "correspondence"

    @classmethod
    def label(cls) -> Str:
        return Str._("Correspondence")


class Confirmation(DuringLifeEventType):
    @classmethod
    def name(cls) -> str:
        return "confirmation"

    @classmethod
    def label(cls) -> Str:
        return Str._("Confirmation")


class Missing(DuringLifeEventType):
    @classmethod
    def name(cls) -> str:
        return "missing"

    @classmethod
    def label(cls) -> Str:
        return Str._("Missing")


class Conference(DuringLifeEventType):
    @classmethod
    def name(cls) -> str:
        return "conference"

    @classmethod
    def label(cls) -> Str:
        return Str._("Conference")



"""
Integrate Betty with `Webpack <https://webpack.js.org/>`_.

This module is internal.
"""

from __future__ import annotations

from asyncio import to_thread
from collections.abc import Callable, Sequence
from pathlib import Path
from shutil import copytree
from typing import Any

from aiofiles.tempfile import TemporaryDirectory

from betty import fs
from betty._npm import NpmRequirement, NpmUnavailable
from betty.app.extension import Extension
from betty.extension.webpack import build
from betty.extension.webpack.build import webpack_build_id
from betty.extension.webpack.jinja2.filter import FILTERS
from betty.generate import Generator, GenerationContext
from betty.html import CssProvider
from betty.jinja2 import Jinja2Provider
from betty.job import Context
from betty.locale import Str
from betty.requirement import (
    Requirement,
    AllRequirements,
    AnyRequirement,
    RequirementError,
)


def _prebuilt_webpack_build_directory_path(
    entrypoint_providers: Sequence[WebpackEntrypointProvider & Extension],
) -> Path:
    return (
        fs.PREBUILT_ASSETS_DIRECTORY_PATH
        / "webpack"
        / f"build-{webpack_build_id(entrypoint_providers)}"
    )


class WebpackEntrypointProvider:
    @classmethod
    def webpack_entrypoint_directory_path(cls) -> Path:
        """
        Get the path to the directory with the entrypoint assets.

        The directory must include at least a ``package.json`` and ``main.ts``.
        """
        raise NotImplementedError

    def webpack_entrypoint_cache_keys(self) -> Sequence[str]:
        """
        Get the keys that make a Webpack build for this provider unique.

        Providers that can be cached regardless may ``return ()``.
        """
        raise NotImplementedError


class PrebuiltAssetsRequirement(Requirement):
    def is_met(self) -> bool:
        return (fs.PREBUILT_ASSETS_DIRECTORY_PATH / "webpack").is_dir()

    def summary(self) -> Str:
        return (
            Str._("Pre-built Webpack front-end assets are available")
            if self.is_met()
            else Str._("Pre-built Webpack front-end assets are unavailable")
        )


class Webpack(Extension, CssProvider, Jinja2Provider, Generator):
    _npm_requirement = NpmRequirement()
    _prebuilt_assets_requirement = PrebuiltAssetsRequirement()
    _requirement = AnyRequirement(
        _npm_requirement,
        _prebuilt_assets_requirement,
    )

    @classmethod
    def name(cls) -> str:
        return "betty.extension.Webpack"

    @classmethod
    def enable_requirement(cls) -> Requirement:
        return AllRequirements(super().enable_requirement(), cls._requirement)

    def build_requirement(self) -> Requirement:
        return self._npm_requirement

    @classmethod
    def assets_directory_path(cls) -> Path:
        return Path(__file__).parent / "assets"

    @property
    def public_css_paths(self) -> list[str]:
        return [
            self.app.static_url_generator.generate("css/vendor.css"),
        ]

    def new_context_vars(self) -> dict[str, Any]:
        return {
            "webpack_js_entrypoints": set(),
        }

    @property
    def filters(self) -> dict[str, Callable[..., Any]]:
        return FILTERS

    @property
    def _project_entrypoint_providers(
        self,
    ) -> Sequence[WebpackEntrypointProvider & Extension]:
        return [
            extension
            for extension in self._app.extensions.flatten()
            if isinstance(extension, WebpackEntrypointProvider)
        ]

    async def generate(self, job_context: GenerationContext) -> None:
        build_directory_path = await self._generate_ensure_build_directory(
            job_context=job_context,
        )
        await self._copy_build_directory(
            build_directory_path, self._app.project.configuration.www_directory_path
        )

    async def prebuild(self, job_context: Context) -> None:
        async with TemporaryDirectory() as working_directory_path_str:
            build_directory_path = await self._new_builder(
                Path(working_directory_path_str),
                job_context=job_context,
            ).build()
            await self._copy_build_directory(
                build_directory_path,
                _prebuilt_webpack_build_directory_path(
                    self._project_entrypoint_providers
                ),
            )

    def _new_builder(
        self,
        working_directory_path: Path,
        *,
        job_context: Context,
    ) -> build.Builder:
        return build.Builder(
            working_directory_path,
            self._project_entrypoint_providers,
            self._app.project.configuration.debug,
            self._app.renderer,
            job_context=job_context,
            localizer=self._app.localizer,
        )

    async def _copy_build_directory(
        self,
        build_directory_path: Path,
        destination_directory_path: Path,
    ) -> None:
        await to_thread(
            copytree,
            build_directory_path,
            destination_directory_path,
            dirs_exist_ok=True,
        )

    async def _generate_ensure_build_directory(
        self,
        *,
        job_context: Context,
    ) -> Path:
        builder = self._new_builder(
            self._app.binary_file_cache.with_scope("webpack").path,
            job_context=job_context,
        )
        try:
            # (Re)build the assets if `npm` is available.
            return await builder.build()
        except NpmUnavailable:
            pass

        # Use prebuilt assets if they exist.
        prebuilt_webpack_build_directory_path = _prebuilt_webpack_build_directory_path(
            self._project_entrypoint_providers
        )
        if prebuilt_webpack_build_directory_path.exists():
            return prebuilt_webpack_build_directory_path

        raise RequirementError(self._requirement)
